/*
 * xkeymacs substitute for X11/Linux
 * Copyright 2011 (C) Takeshi Yashiro <yashiromann@nifty.com>
 *
 * emacs.h: keyboard input converter (Emacs keybinding)
 */

#include <stdio.h>
#include "window.h"
#include "emacs.h"

#define DP(x) if (m_debug) printf x

#define BIT_LEFTCTRL (1 << 0)
#define BIT_RIGHTCTRL (1 << 1)
#define BIT_LEFTSHIFT (1 << 2)
#define BIT_RIGHTSHIFT (1 << 3)
#define BIT_LEFTALT (1 << 4)
#define BIT_RIGHTALT (1 << 5)
#define IS_CTRL_ON() ((m_metaKeyFlags & (BIT_LEFTCTRL|BIT_RIGHTCTRL)) != 0)
#define IS_SHIFT_ON() ((m_metaKeyFlags & (BIT_LEFTSHIFT|BIT_RIGHTSHIFT)) != 0)
#define IS_ALT_ON() ((m_metaKeyFlags & (BIT_LEFTALT|BIT_RIGHTALT)) != 0)

EmacsConverter::EmacsConverter(const std::vector<std::string>& excludeList, bool debug)
    : Converter(debug), m_metaKeyFlags(0), 
      m_state_ctrlx(false), m_state_ctrlspc(false), 
      m_state_ctrlk(false), m_state_pasteNewline(false),
      m_debug(debug)
{
    std::vector<std::string>::const_iterator iend = excludeList.end();
    for (std::vector<std::string>::const_iterator i = excludeList.begin();
         i != iend; ++i) {
        if (i->size() <= 0)
            continue;
        addExcludePattern(i->c_str());
    }
}

EmacsConverter::~EmacsConverter()
{
    ExcludeList::const_iterator iend = m_excludePatterns.end();
    for (ExcludeList::iterator i = m_excludePatterns.begin();
         i != iend; ++i) {
        delete (*i);
    }
}

void EmacsConverter::addExcludePattern(const char* pattern)
{
    m_excludePatterns.push_back(new boost::regex(pattern));
}

bool EmacsConverter::handleMetaKeyInput(struct input_event* input, int bit)
{
    if (input->value == 1)
        m_metaKeyFlags |= bit;
    else if (input->value == 0)
        m_metaKeyFlags &= ~bit;
    addOutput(input);
    return true;
}

void EmacsConverter::typeKey(__u16 code, int metaKeys)
{
    int tmpRelease = m_metaKeyFlags & ~metaKeys;
    int tmpPress = ~m_metaKeyFlags & metaKeys;
    static const int bits[] = {
        BIT_LEFTCTRL, BIT_RIGHTCTRL, 
        BIT_LEFTSHIFT, BIT_RIGHTSHIFT, 
        BIT_LEFTALT, BIT_RIGHTALT,
    };
    static const __u16 keys[] = {
        KEY_LEFTCTRL, KEY_RIGHTCTRL, 
        KEY_LEFTSHIFT, KEY_RIGHTSHIFT, 
        KEY_LEFTALT, KEY_RIGHTALT,
    };

    struct input_event input, syn;
    memset(&input, 0, sizeof(input));
    input.type = EV_KEY;
    memset(&syn, 0, sizeof(syn));
    syn.type = EV_SYN;
    syn.code = SYN_REPORT;
    syn.value = 0;

    /* Change meta-key states to metaKeys */
    input.value = 0;
    for (int i = 0; i < int(sizeof(keys)/sizeof(keys[0])); i++) {
        if (tmpRelease & bits[i]) {
            input.code = keys[i];
            addOutput(&input);
        }
    }

    input.value = 1;
    for (int i = 0; i < int(sizeof(keys)/sizeof(keys[0])); i++) {
        if (tmpPress & bits[i]) {
            input.code = keys[i];
            addOutput(&input);
        }
    }

    addOutput(&syn);

    /* Press & release the given keycode */
    input.code = code;

    input.value = 1;
    addOutput(&input);
    addOutput(&syn);

    input.value = 0;
    addOutput(&input);
    addOutput(&syn);

    /* Change meta-key states back to m_metaKeyFlags */
    input.value = 0;
    for (int i = 0; i < int(sizeof(keys)/sizeof(keys[0])); i++) {
        if (tmpPress & bits[i]) {
            input.code = keys[i];
            addOutput(&input);
        }
    }

    input.value = 1;
    for (int i = 0; i < int(sizeof(keys)/sizeof(keys[0])); i++) {
        if (tmpRelease & bits[i]) {
            input.code = keys[i];
            addOutput(&input);
        }
    }

    addOutput(&syn);
}

bool EmacsConverter::handleKeyInput(struct input_event* input)
{
    switch (input->code) {
    case KEY_LEFTCTRL:
        return handleMetaKeyInput(input, BIT_LEFTCTRL);
    case KEY_RIGHTCTRL:
        return handleMetaKeyInput(input, BIT_RIGHTCTRL);
    case KEY_LEFTSHIFT:
        return handleMetaKeyInput(input, BIT_LEFTSHIFT);
    case KEY_RIGHTSHIFT:
        return handleMetaKeyInput(input, BIT_RIGHTSHIFT);
    case KEY_LEFTALT:
        return handleMetaKeyInput(input, BIT_LEFTALT);
    case KEY_RIGHTALT:
        return handleMetaKeyInput(input, BIT_RIGHTALT);
    case KEY_X:
        if (!exclude() && input->value > 0
            && IS_CTRL_ON() && !IS_SHIFT_ON() && !IS_ALT_ON()) {
            /* C-x */
            m_state_ctrlx = true;
            return true;
        }
        handleKeyDefault(input);
        return true;
    case KEY_F:
        if (!exclude() && input->value > 0 
            && IS_CTRL_ON() && !IS_SHIFT_ON() && !IS_ALT_ON()) {
            if (m_state_ctrlx) {
                /* C-x C-f */
                DP(("C-x C-f\n"));
                typeKey(KEY_O, BIT_LEFTCTRL);
                m_state_ctrlx = false;
                m_state_ctrlk = false;
                return true;
            }
            else {
                /* C-f */
                DP(("C-f\n"));
                typeKey(KEY_RIGHT, 
                        (m_state_ctrlspc ? BIT_LEFTSHIFT : 0));
                m_state_ctrlk = false;
                return true;
            }
        }
        handleKeyDefault(input);
        return true;
    case KEY_C:
        if (!exclude() && input->value > 0
            && IS_CTRL_ON() && !IS_SHIFT_ON() && !IS_ALT_ON()) {
            if (m_state_ctrlx) {
                /* C-x C-c */
                DP(("C-x C-c\n"));
                typeKey(KEY_F4, BIT_LEFTALT);
                m_state_ctrlx = false;
                m_state_ctrlk = false;
                return true;
            }
        }
        handleKeyDefault(input);
        return true;
    case KEY_D:
        if (!exclude() && input->value > 0
            && IS_CTRL_ON() && !IS_SHIFT_ON() && !IS_ALT_ON()) {
            /* C-d */
            DP(("C-d\n"));
            typeKey(KEY_DELETE, 0);
            m_state_ctrlspc = false;
            m_state_ctrlk = false;
            return true;
        }
        handleKeyDefault(input);
        return true;
    case KEY_H:
        if (!exclude() && input->value > 0) {
            if (IS_CTRL_ON() && !IS_SHIFT_ON() && !IS_ALT_ON()) {
                /* C-h */
                DP(("C-h\n"));
                typeKey(KEY_BACKSPACE, 0);
                m_state_ctrlspc = false;
                m_state_ctrlk = false;
                return true;
            }
            else if (m_state_ctrlx && 
                     !IS_CTRL_ON() && !IS_SHIFT_ON() && !IS_ALT_ON()) {
                /* C-x h */
                DP(("C-x h\n"));
                typeKey(KEY_A, BIT_LEFTCTRL);
                m_state_ctrlx = false;
                m_state_ctrlk = false;
                return true;
            }
        }
        handleKeyDefault(input);
        return true;
    case KEY_K:
        if (!exclude() && input->value > 0) {
            if (IS_CTRL_ON() && !IS_SHIFT_ON() && !IS_ALT_ON()) {
                /* C-k */
                DP(("C-k\n"));
                if (m_state_ctrlk) {
                    typeKey(KEY_DELETE, 0);
                    m_state_pasteNewline = true;
                }
                else {
                    typeKey(KEY_END, BIT_LEFTSHIFT);
                    struct timespec ts;
                    ts.tv_sec = 0;
                    ts.tv_nsec = 10 * 1000 * 1000; // 10ms
                    nanosleep(&ts, NULL);
                    typeKey(KEY_X, BIT_LEFTCTRL);
                    m_state_pasteNewline = false;
                }
                m_state_ctrlspc = false;
                m_state_ctrlk = !m_state_ctrlk;
                return true;
            }
            else if (m_state_ctrlx && 
                     !IS_CTRL_ON() && !IS_SHIFT_ON() && !IS_ALT_ON()) {
                /* C-x k */
                DP(("C-x k\n"));
                typeKey(KEY_W, BIT_LEFTCTRL);
                m_state_ctrlx = false;
                m_state_ctrlk = false;
                return true;
            }
        }
        handleKeyDefault(input);
        return true;
    case KEY_O:
        if (!exclude() && input->value > 0
            && IS_CTRL_ON() && !IS_SHIFT_ON() && !IS_ALT_ON()) {
            /* C-o */
            DP(("C-o\n"));
            typeKey(KEY_END, 0);
            typeKey(KEY_ENTER, 0);
            typeKey(KEY_UP, 0);
            m_state_ctrlspc = false;
            m_state_ctrlk = false;
            return true;
        }
        handleKeyDefault(input);
        return true;
    case KEY_G:
        if (!exclude() && input->value > 0
            && IS_CTRL_ON() && !IS_SHIFT_ON() && !IS_ALT_ON()) {
            /* C-g */
            DP(("C-g\n"));
            typeKey(KEY_ESC, 0);
            m_state_ctrlspc = false;
            m_state_ctrlx = false;
            m_state_ctrlk = false;
            return true;
        }
        handleKeyDefault(input);
        return true;
    case KEY_J:
        if (!exclude() && input->value > 0
            && IS_CTRL_ON() && !IS_SHIFT_ON() && !IS_ALT_ON()) {
            /* C-j */
            DP(("C-j\n"));
            typeKey(KEY_ENTER, 0);
            typeKey(KEY_TAB, 0);
            m_state_ctrlspc = false;
            m_state_ctrlk = false;
            return true;
        }
        handleKeyDefault(input);
        return true;
    case KEY_M:
        if (!exclude() && input->value > 0
            && IS_CTRL_ON() && !IS_SHIFT_ON() && !IS_ALT_ON()) {
            /* C-m */
            DP(("C-m\n"));
            typeKey(KEY_ENTER, 0);
            m_state_ctrlspc = false;
            m_state_ctrlk = false;
            return true;
        }
        handleKeyDefault(input);
        return true;
    case KEY_I:
        if (!exclude() && input->value > 0
            && IS_CTRL_ON() && !IS_SHIFT_ON() && !IS_ALT_ON()) {
            /* C-i */
            DP(("C-i\n"));
            typeKey(KEY_TAB, 0);
            m_state_ctrlspc = false;
            m_state_ctrlk = false;
            return true;
        }
        handleKeyDefault(input);
        return true;
    case KEY_S:
        if (!exclude() && input->value > 0
            && IS_CTRL_ON() && !IS_SHIFT_ON() && !IS_ALT_ON()) {
            if (m_state_ctrlx) {
                /* C-x C-s */
                DP(("C-x C-s\n"));
                typeKey(KEY_S, BIT_LEFTCTRL);
                m_state_ctrlx = false;
                m_state_ctrlk = false;
                return true;
            }
            else {
                /* C-s */
                DP(("C-s\n"));
                typeKey(KEY_F, BIT_LEFTCTRL);
                m_state_ctrlspc = false;
                m_state_ctrlk = false;
                return true;
            }
        }
        handleKeyDefault(input);
        return true;
    case KEY_R:
        if (!exclude() && input->value > 0
            && IS_CTRL_ON() && !IS_SHIFT_ON() && !IS_ALT_ON()) {
            /* C-r */
            DP(("C-r\n"));
            typeKey(KEY_F, BIT_LEFTCTRL);
            m_state_ctrlspc = false;
            m_state_ctrlk = false;
            return true;
        }
        handleKeyDefault(input);
        return true;
    case KEY_W:
        if (!exclude() && input->value > 0) {
            if (IS_CTRL_ON() && !IS_SHIFT_ON() && !IS_ALT_ON()) {
                /* C-w */
                DP(("C-w\n"));
                typeKey(KEY_X, BIT_LEFTCTRL);
                m_state_ctrlspc = false;
                m_state_ctrlk = false;
                m_state_pasteNewline = false;
                return true;
            }
            else if (!IS_CTRL_ON() && !IS_SHIFT_ON() && IS_ALT_ON()) {
                /* M-w */
                DP(("M-w\n"));
                typeKey(KEY_C, BIT_LEFTCTRL);
                m_state_ctrlspc = false;
                m_state_ctrlk = false;
                m_state_pasteNewline = false;
                return true;
            }
        }
        handleKeyDefault(input);
        return true;
    case KEY_Y:
        if (!exclude() && input->value > 0
            && IS_CTRL_ON() && !IS_SHIFT_ON() && !IS_ALT_ON()) {
            /* C-y */
            DP(("C-y\n"));
            typeKey(KEY_V, BIT_LEFTCTRL);
            if (m_state_pasteNewline)
                typeKey(KEY_ENTER, 0);
            m_state_ctrlspc = false;
            m_state_ctrlk = false;
            return true;
        }
        handleKeyDefault(input);
        return true;
    case KEY_SLASH:
        if (!exclude() && input->value > 0
            && IS_CTRL_ON() && !IS_SHIFT_ON() && !IS_ALT_ON()) {
            /* C-/ */
            DP(("C-/\n"));
            typeKey(KEY_Z, BIT_LEFTCTRL);
            m_state_ctrlspc = false;
            m_state_ctrlk = false;
            return true;
        }
        handleKeyDefault(input);
        return true;
#if 1
    /* On JP-106/109 Keyboard, `@' key seems to be assigned to KEY_LEFTBRACE */ 
    case KEY_LEFTBRACE:
#endif
    case KEY_SPACE:
        if (!exclude() && input->value > 0
            && IS_CTRL_ON() && !IS_SHIFT_ON() && !IS_ALT_ON()) {
            /* C-SPC */
            DP(("C-SPC\n"));
            m_state_ctrlspc = !m_state_ctrlspc;
            m_state_ctrlk = false;
            return true;
        }
        handleKeyDefault(input);
        return true;
    case KEY_A:
        if (!exclude() && input->value > 0
            && IS_CTRL_ON() && !IS_SHIFT_ON() && !IS_ALT_ON()) {
            /* C-a */
            DP(("C-a\n"));
            typeKey(KEY_HOME, 
                    (m_state_ctrlspc ? BIT_LEFTSHIFT : 0));
            m_state_ctrlk = false;
            return true;
        }
        handleKeyDefault(input);
        return true;        
    case KEY_E:
        if (!exclude() && input->value > 0
            && IS_CTRL_ON() && !IS_SHIFT_ON() && !IS_ALT_ON()) {
            /* C-e */
            DP(("C-e\n"));
            typeKey(KEY_END, 
                    (m_state_ctrlspc ? BIT_LEFTSHIFT : 0));
            m_state_ctrlk = true;
            return true;
        }
        handleKeyDefault(input);
        return true;        
    case KEY_P:
        if (!exclude() && input->value > 0
            && IS_CTRL_ON() && !IS_SHIFT_ON() && !IS_ALT_ON()) {
            /* C-p */
            DP(("C-p\n"));
            typeKey(KEY_UP, 
                    (m_state_ctrlspc ? BIT_LEFTSHIFT : 0));
            m_state_ctrlk = false;
            return true;
        }
        handleKeyDefault(input);
        return true;
    case KEY_N:
        if (!exclude() && input->value > 0
            && IS_CTRL_ON() && !IS_SHIFT_ON() && !IS_ALT_ON()) {
            /* C-p */
            DP(("C-n\n"));
            typeKey(KEY_DOWN, 
                    (m_state_ctrlspc ? BIT_LEFTSHIFT : 0));
            m_state_ctrlk = false;
            return true;
        }
        handleKeyDefault(input);
        return true;
    case KEY_B:
        if (!exclude() && input->value > 0
            && IS_CTRL_ON() && !IS_SHIFT_ON() && !IS_ALT_ON()) {
            /* C-b */
            DP(("C-b\n"));
            typeKey(KEY_LEFT, 
                    (m_state_ctrlspc ? BIT_LEFTSHIFT : 0));
            m_state_ctrlk = false;
            return true;
        }
        handleKeyDefault(input);
        return true;
    case KEY_V:
        if (!exclude() && input->value > 0) {
            if (IS_CTRL_ON() && !IS_SHIFT_ON() && !IS_ALT_ON()) {
                DP(("C-v\n"));
                typeKey(KEY_PAGEDOWN, 
                        (m_state_ctrlspc ? BIT_LEFTSHIFT : 0));
                m_state_ctrlk = false;
                return true;
            }
            else if (!IS_CTRL_ON() && !IS_SHIFT_ON() && IS_ALT_ON()) {
                DP(("M-v\n"));
                typeKey(KEY_PAGEUP, 
                        (m_state_ctrlspc ? BIT_LEFTSHIFT : 0));
                m_state_ctrlk = false;
                return true;
            }
        }
        handleKeyDefault(input);
        return true;
    default:
        handleKeyDefault(input);
        return true;
    }
}

bool EmacsConverter::handleInput(struct input_event* input)
{
    switch (input->type) {
    case EV_KEY:
        /* Key events are manipulated */
        return handleKeyInput(input);
    default:
        /* For other events, simply send the event as it is */
        addOutput(input);
        return true;
    }
}

bool EmacsConverter::handleKeyDefault(struct input_event* input)
{
    addOutput(input);
    if (input->value == 1)
        m_state_ctrlk = false;
    return true;
}

bool EmacsConverter::exclude() const
{
    std::string winTitle = 
        WindowSystem::getInstance()->getFocusedWindowTitle();
    DP(("winTitle: \"%s\"\n", winTitle.c_str()));

    ExcludeList::const_iterator iend = m_excludePatterns.end();
    for (ExcludeList::const_iterator i = m_excludePatterns.begin();
         i != iend; ++i) {
        if (boost::regex_search(winTitle, **i))
            return true;
    }
    return false;
}
