/*
 * xkeymacs substitute for X11/Linux
 * Copyright 2011 (C) Takeshi Yashiro <yashiromann@nifty.com>
 *
 * main.cpp: main routine
 */

#include <unistd.h>
#include <termios.h>
#include <stdio.h>
#include <string.h>
#include <memory>
#include <string>
#include <vector>
#include "keyboard.h"
#include "conv.h"
#include "emacs.h"
#include "window.h"

static void usage()
{
    fprintf(stderr, "Usage: x11keymacs <options> /dev/input/event<n>\n");
    fprintf(stderr, "  -e|--exclude <file>: exclude window titles matching patterns in <file>\n");
    fprintf(stderr, "  -n|--no-grab: do not grab device input (for debug)\n");
    fprintf(stderr, "  -d|--debug: enable debug mode\n");
}

int main(int argc, char* argv[])
{
    char* fileName = NULL;
    bool noGrabMode = false;
    bool debugMode = false;
    const char* excludeFile = NULL;
    std::vector<std::string> excludeList;

    /* Parse command line arguments */
    for (int i = 1; i < argc; i++) {
        if (argv[i][0] == '-') {
            if (strcmp(argv[i], "-e") == 0 
                || strcmp(argv[i], "--exclude") == 0) {
                if (i+1 >= argc || argv[i+1][0] == '-') {
                    fprintf(stderr, "ERROR: option `%s' needs one argument\n",
                            argv[i]);
                    usage();
                    return 1;
                }
                excludeFile = argv[++i];
            }
            else if (strcmp(argv[i], "-n") == 0 
                || strcmp(argv[i], "--no-grab") == 0) {
                noGrabMode = true;
            }
            else if (strcmp(argv[i], "-d") == 0 
                || strcmp(argv[i], "--debug") == 0) {
                debugMode = true;
            }
            else {
                fprintf(stderr, "ERROR: bad option `%s'\n", argv[i]);
                usage();
                return 1;
            }
        }
        else {
            if (fileName != NULL) {
                fprintf(stderr, "ERROR: multiple targets given\n");
                usage();
                return 1;
            }
            fileName = argv[i];
        }
    }
    if (fileName == NULL) {
        fprintf(stderr, "ERROR: device file not specified\n");
        usage();
        return 1;
    }
    if (!WindowSystem::getInstance()->isValid()) {
        fprintf(stderr, "ERROR: cannot open display\n");
        return 1;
    }

    /* Read exclude list */
    if (excludeFile != NULL) {
        FILE *fp = fopen(excludeFile, "r");
        if (fp == NULL) {
            fprintf(stderr, "ERROR: failed to open file `%s'\n", excludeFile);
            return 1;
        }
        char buf[256];
        while (fgets(buf, sizeof(buf), fp) == buf) {
            for (int i = strlen(buf) - 1; i >= 0; i--) {
                if (buf[i] != '\n' && buf[i] != '\r')
                    break;
                buf[i] = '\0';
            }
            excludeList.push_back(std::string(buf));
        }
        fclose(fp);
    }

    /* We don't want keyboard echos */
    struct termios term;
    if (tcgetattr(STDIN_FILENO, &term) >= 0) {
        term.c_lflag &= ~ECHO;
        tcsetattr(STDIN_FILENO, TCSADRAIN, &term);
    }

    /* Initialize keyboard device */
    std::auto_ptr<KeyboardDevice> 
        device(new KeyboardDevice(fileName, !noGrabMode, debugMode));
    if (!device->isKeyDevValid()) {
        fprintf(stderr, "ERROR: cannot open device `%s\'\n", fileName);
        return 1;
    }
    if (!device->isUserDevValid()) {
        fprintf(stderr, "ERROR: device open device `%s\'\n", USERDEV_FILENAME);
        return 1;
    }

    /* Initialize converter */
    std::auto_ptr<Converter>
        conv(new EmacsConverter(excludeList, debugMode));

    /* Get event, convert, and put the generated events */
    for (;;) {
        struct input_event key;
        device->getKey(&key);
        conv->addInput(&key);
        while (conv->hasOutput()) {
            conv->getOutput(&key);
            device->putKey(&key);
        }
    }

    return 0;
}
